import os
import re
import vcr
import pytest
from opperai import AsyncClient, Client


@pytest.fixture(scope="module")
def aclient() -> AsyncClient:
    yield AsyncClient()


@pytest.fixture(scope="module")
def client() -> Client:
    yield Client()


def uri_matcher(r1, r2):
    url_pattern = re.compile(r"https?://(api\.opper\.ai|localhost:8000)")
    uri1 = re.sub(url_pattern, "", r1.uri)
    uri2 = re.sub(url_pattern, "", r2.uri)

    return uri1 == uri2


@pytest.fixture
def vcr_cassette(request):
    test_name = request.node.name
    module_file = request.module.__file__
    file_name = os.path.splitext(os.path.basename(module_file))[0]

    cassette_name = f"{file_name}/{test_name}.yaml"

    my_vcr = vcr.VCR(
        cassette_library_dir="tests/fixtures/vcr_cassettes",
        path_transformer=vcr.VCR.ensure_suffix(".yaml"),
        filter_headers=[
            "x-opper-api-key",
            "host",
        ],
    )
    my_vcr.register_matcher("uri_matcher", uri_matcher)
    my_vcr.match_on = ["method", "body", "headers", "uri_matcher"]

    with my_vcr.use_cassette(cassette_name):
        yield



from unittest.mock import MagicMock, patch

import pytest
import random
from opperai import Client
from opperai.types import ChatPayload, FunctionDescription, Message, CacheConfiguration

api_key = "key"


@patch("opperai._http_clients._http_client.do_request")
def test_chat(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200, json=lambda: {"message": "Bonjour"}
    )
    client = Client(api_key=api_key)
    resp = client.functions.chat(
        "french", ChatPayload(messages=[Message(role="user", content="hello")])
    )

    assert resp.message == "Bonjour"
    mock_do_request.assert_called_once()


@pytest.mark.asyncio
@patch("opperai._http_clients._http_client.stream")
async def test_chat_stream(mock_stream):
    def gen():
        yield {"delta": "Bon"}
        yield {"delta": "Jour"}

    mock_stream.return_value = gen()
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    gen = client.functions.chat(
        "french",
        ChatPayload(messages=[Message(role="user", content="hello")]),
        stream=True,
    )
    resp = ""
    for message in gen:
        resp += message.delta

    assert resp == "BonJour"


@pytest.mark.asyncio
@patch("opperai._http_clients._http_client.do_request")
async def test_create_function(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(
            status_code=404,
        ),  # Response for get_function_by_path
        MagicMock(status_code=200, json=lambda: {"id": 1}),  # Response for create
    ]
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    function = FunctionDescription(
        path="test/path", description="Test function", instructions="Do something"
    )
    fn_id = client.functions.create(function)
    assert fn_id == 1
    assert mock_do_request.call_count == 2
    mock_do_request.assert_any_call(
        "GET",
        "/api/v1/functions/by_path/test/path",
    )
    mock_do_request.assert_any_call(
        "POST",
        "/api/v1/functions",
        json=function.model_dump(),
    )


@pytest.mark.asyncio
@patch("opperai._http_clients._http_client.do_request")
async def test_update_function(mock_do_request):
    mock_do_request.return_value = MagicMock(status_code=200, json=lambda: {"id": 1})
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    function = FunctionDescription(
        id=1,
        path="test/path",
        description="Updated Test function",
        instructions="Do something else",
    )
    fn_id = client.functions.update(function)
    assert fn_id == function.id
    mock_do_request.assert_called_once()


@pytest.mark.asyncio
@patch("opperai._http_clients._http_client.do_request")
async def test_get(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(
            status_code=200,
            json=lambda: {
                "id": 1,
                "path": "test/path",
                "description": "Test function",
                "instructions": "Do something",
            },
        ),
        MagicMock(
            status_code=200,
            json=lambda: {
                "id": 1,
                "path": "test/path",
                "description": "Test function",
                "instructions": "Do something",
            },
        ),
    ]
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    function_description = client.functions.get(path="test/path")
    assert function_description.id == 1
    assert function_description.path == "test/path"
    assert function_description.description == "Test function"
    mock_do_request.assert_called_once_with(
        "GET",
        "/api/v1/functions/by_path/test/path",
    )

    mock_do_request.reset_mock()

    function_description = client.functions.get(id=1)
    assert function_description.id == 1
    assert function_description.path == "test/path"
    assert function_description.description == "Test function"
    mock_do_request.assert_called_once_with(
        "GET",
        "/api/v1/functions/1",
    )


@pytest.mark.asyncio
@patch("opperai._http_clients._http_client.do_request")
async def test_get_function_by_path(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: {
            "id": 1,
            "path": "test/path",
            "description": "Test function",
            "instructions": "Do something",
        },
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    function_description = client.functions.get_by_path("test/path")
    assert function_description.id == 1
    assert function_description.path == "test/path"
    assert function_description.description == "Test function"
    mock_do_request.assert_called_once_with(
        "GET",
        "/api/v1/functions/by_path/test/path",
    )


@pytest.mark.asyncio
@patch("opperai._http_clients._http_client.do_request")
async def test_get_function_by_id(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: {
            "id": 1,
            "path": "test/path",
            "description": "Test function",
            "instructions": "Do something",
        },
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    function_description = client.functions.get_by_id("1")
    assert function_description.id == 1
    assert function_description.path == "test/path"
    assert function_description.description == "Test function"
    mock_do_request.assert_called_once_with(
        "GET",
        "/api/v1/functions/1",
    )


def test_create_function_with_cache(client: Client, vcr_cassette):
    fdesc = FunctionDescription(
        path="test/sdk/test_create_function_with_cache_sync",
        description="Test function",
        instructions="Do something",
        cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
    )

    fid = client.functions.create(fdesc)
    assert fid is not None

    res = client.functions.chat(
        fdesc.path, ChatPayload(messages=[Message(role="user", content="hello")])
    )
    print(res)
    assert not res.cached

    res = client.functions.chat(
        fdesc.path, ChatPayload(messages=[Message(role="user", content="hello")])
    )
    print(res)
    assert res.cached

    client.functions.delete(path=fdesc.path)



import os
import tempfile
from unittest.mock import ANY, MagicMock, call, patch

from opperai import Client
from opperai.types.indexes import DocumentIn, Filter


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_retrieve(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: [
            {
                "score": 2.3,
                "id": 1,
                "content": "Bonjour",
                "metadata": {"source": "test"},
            }
        ],
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    resp = client.indexes.retrieve(index_id=1, query="test", k=1)

    assert resp[0].content == "Bonjour"
    assert resp[0].metadata == {"source": "test"}
    mock_do_request.assert_called_once()


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_retrieve_filters(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: [
            {
                "score": 2.3,
                "id": 1,
                "content": "Bonjour",
                "metadata": {"source": "test"},
            }
        ],
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    resp = client.indexes.retrieve(
        index_id=1,
        query="test",
        k=1,
        filters=[Filter(key="source", operation="=", value="test")],
    )

    assert resp[0].content == "Bonjour"
    assert resp[0].metadata == {"source": "test"}
    mock_do_request.assert_called_once()


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_create(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: {"id": 123},
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    index_id = client.indexes.create(name="test_index")

    assert index_id == 123
    mock_do_request.assert_called_once_with(
        "POST",
        "/v1/indexes",
        json={"name": "test_index"},
    )


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_index(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: {
            "id": 1,
            "uuid": "506768ec-54ac-4233-8a7b-ea9d1f12b879",
            "key": "test_key",
        },
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    doc_in = DocumentIn(content="Hello", metadata={"source": "test"})
    doc_out = client.indexes.index(index_id=42, doc=doc_in)

    assert doc_out.id == 1
    assert str(doc_out.uuid) == "506768ec-54ac-4233-8a7b-ea9d1f12b879"
    assert doc_out.key == "test_key"
    mock_do_request.assert_called_once()


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_list(mock_do_request):
    mock_do_request.return_value = MagicMock(
        status_code=200,
        json=lambda: [
            {"id": 1, "name": "test_index_1", "created_at": "2021-07-21T17:32:28Z"},
            {"id": 2, "name": "test_index_2", "created_at": "2021-07-22T17:32:28Z"},
        ],
    )
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    indexes = client.indexes.list()

    assert len(indexes) == 2
    assert indexes[0].id == 1
    assert indexes[0].name == "test_index_1"
    assert indexes[1].id == 2
    assert indexes[1].name == "test_index_2"
    mock_do_request.assert_called_once_with(
        "GET",
        "/v1/indexes",
    )


@patch("opperai._http_clients._http_client.do_request")
def test_indexes_upload_file(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(
            status_code=200,
            json=lambda: {
                "url": "http://localhost:8000/upload",
                "fields": {"key": "value"},
                "uuid": "123e4567-e89b-12d3-a456-426614174000",
            },
        ),
        MagicMock(status_code=204),
        MagicMock(
            status_code=200,
            json=lambda: {"uuid": "123e4567-e89b-12d3-a456-426614174000"},
        ),
    ]
    client = Client(api_key="op-dev-api-key", api_url="http://localhost:8000")
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_file:
        tmp_file.write("Some test content")
        tmp_file_path = tmp_file.name
    response = client.indexes.upload_file(index_id=1, file_path=tmp_file_path)
    os.remove(tmp_file_path)  # Clean up the temporary file after use

    assert response == {"uuid": "123e4567-e89b-12d3-a456-426614174000"}
    mock_do_request.assert_has_calls(
        [
            call("GET", "/v1/indexes/1/upload_url", params={"filename": ANY}),
            call(
                "POST", "http://localhost:8000/upload", files=ANY, data={"key": "value"}
            ),
            call(
                "POST",
                "/v1/indexes/1/register_file",
                json={"uuid": "123e4567-e89b-12d3-a456-426614174000"},
            ),
        ]
    )



import asyncio
import contextvars
import datetime
import json
import os
from contextlib import contextmanager
from functools import wraps
from typing import Callable, Optional
from uuid import uuid4

from opperai import Client
from opperai.types.spans import Span
from opperai.utils import convert_function_call_to_json

_current_span_id = contextvars.ContextVar("_current_span_id", default=None)


class SpanContext:
    def __init__(self, client: Client, span_uuid: str):
        self.client = client
        self.span_uuid = span_uuid
        self.output = None


@contextmanager
def start_span(
    name: str,
    input: Optional[str] = None,
    metadata: Optional[dict] = None,
    client: Optional[Client] = None,
):
    c = client if client is not None else Client()
    project = os.environ.get("OPPER_PROJECT", "missing_project")
    parent_span_id = _current_span_id.get()
    span_id = str(uuid4())
    span = Span(
        uuid=span_id,
        meta=metadata,
        input=input,
        parent_uuid=parent_span_id if parent_span_id is not None else None,
        project=project,
        name=name,
        start_time=datetime.datetime.utcnow(),
    )
    span_uuid = c.spans.create(span)

    span_token = _current_span_id.set(span_id)
    span_context = SpanContext(c, span_uuid)
    try:
        yield span_context  # This allows the block inside the 'with' statement to execute and interact with the span_context
    finally:
        end_time = datetime.datetime.utcnow()
        update_kwargs = {
            "end_time": end_time,
        }
        if span_context.output is not None:
            update_kwargs["output"] = span_context.output

        c.spans.update(span_uuid, **update_kwargs)
        _current_span_id.reset(span_token)


def trace(
    _func: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    client: Optional[Client] = None,
    trace_io: bool = False,
):
    def decorator_trace(func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            c = client if client is not None else Client()

            project = os.environ.get("OPPER_PROJECT", "missing_project")
            parent_span_id = _current_span_id.get()
            span_id = str(uuid4())
            span_name = name if name is not None else func.__name__
            inputs = None
            if trace_io:
                inputs = json.dumps(
                    convert_function_call_to_json(func, *args, **kwargs)
                )
            span = Span(
                uuid=span_id,
                parent_uuid=parent_span_id if parent_span_id is not None else None,
                project=project,
                name=span_name,
                input=inputs,
                start_time=datetime.datetime.utcnow(),
            )
            span_uuid = c.spans.create(span)

            span_token = _current_span_id.set(span_id)
            try:
                if asyncio.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
                c.spans.update(
                    span_uuid,
                    end_time=datetime.datetime.utcnow(),
                    output=json.dumps(result) if trace_io else None,
                )
            finally:
                _current_span_id.reset(span_token)

            return result

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            c = client if client is not None else Client()
            project = os.environ.get("OPPER_PROJECT", "missing_project")
            parent_span_id = _current_span_id.get()
            span_id = str(uuid4())
            span_name = name if name is not None else func.__name__
            inputs = None
            if trace_io:
                inputs = json.dumps(
                    convert_function_call_to_json(func, *args, **kwargs)
                )
            span = Span(
                uuid=span_id,
                parent_uuid=parent_span_id if parent_span_id is not None else None,
                project=project,
                name=span_name,
                input=inputs,
                start_time=datetime.datetime.utcnow(),
            )
            span_uuid = c.spans.create(span)

            span_token = _current_span_id.set(span_id)
            try:
                result = func(*args, **kwargs)
                c.spans.update(
                    span_uuid,
                    end_time=datetime.datetime.utcnow(),
                    output=json.dumps(result) if trace_io else None,
                )
            finally:
                _current_span_id.reset(span_token)

            return result

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    if _func is None:
        return decorator_trace
    else:
        return decorator_trace(_func)



from typing import Generator

from opperai._http_clients import _async_http_client
from opperai.types import (
    ChatPayload,
    FunctionDescription,
    FunctionResponse,
    StreamingChunk,
    validate_id_xor_path,
)
from opperai.types.exceptions import APIError, RateLimitError


class AsyncFunctions:
    def __init__(self, http_client: _async_http_client, default_model: str = None):
        self.http_client = http_client
        self.default_model = default_model

    async def _create(self, function: FunctionDescription, **kwargs) -> int:
        if not function.model and self.default_model:
            function.model = self.default_model
        response = await self.http_client.do_request(
            "POST",
            "/api/v1/functions",
            json={**function.model_dump(), **kwargs},
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to create function {function.path} with status {response.status_code}"
            )

        return response.json()["id"]

    async def update(self, function: FunctionDescription, **kwargs) -> int:
        response = await self.http_client.do_request(
            "POST",
            f"/api/v1/functions/{function.id}",
            json={**function.model_dump(), **kwargs},
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to update function {function.path} with status {response.status_code}"
            )

        return response.json()["id"]

    @validate_id_xor_path
    async def get(self, id: str = None, path: str = None) -> FunctionDescription:
        if path is not None:
            if id is not None:
                raise ValueError("Only one of id or path should be provided")
            return await self.get_by_path(path)
        elif id is not None:
            return await self.get_by_id(id)
        else:
            return None

    async def get_by_path(self, function_path: str) -> FunctionDescription:
        response = await self.http_client.do_request(
            "GET",
            f"/api/v1/functions/by_path/{function_path}",
        )
        if response.status_code == 404:
            return None
        if response.status_code != 200:
            raise APIError(
                f"Failed to get function {function_path} with status {response.status_code}"
            )

        return FunctionDescription(**response.json())

    async def get_by_id(self, function_id: str) -> FunctionDescription:
        response = await self.http_client.do_request(
            "GET",
            f"/api/v1/functions/{function_id}",
        )
        if response.status_code == 404:
            return None
        if response.status_code != 200:
            raise APIError(
                f"Failed to get function {function_id} with status {response.status_code}"
            )

        return FunctionDescription(**response.json())

    async def create(
        self, function: FunctionDescription, update: bool = True, **kwargs
    ) -> int:
        fn = await self.get_by_path(function.path)
        if fn is None:
            return await self._create(function, **kwargs)
        elif update:
            function.id = fn.id
            return await self.update(function, **kwargs)
        return fn.id

    @validate_id_xor_path
    async def delete(self, id: str = None, path: str = None):
        if path is not None:
            try:
                await self._delete_by_path(path)
            except APIError:
                pass
        elif id is not None:
            fn = await self.get(id=id)
            if fn:
                await self._delete_by_path(fn.path)

    async def _delete_by_path(self, function_path: str):
        response = await self.http_client.do_request(
            "DELETE",
            f"/api/v1/functions/by_path/{function_path}",
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to delete function {function_path} with status {response.status_code}"
            )

    async def chat(
        self, function_path, data: ChatPayload, stream=False, **kwargs
    ) -> FunctionResponse:
        if stream:
            return self._chat_stream(function_path, data, **kwargs)
        serialized_data = data.model_dump()
        response = await self.http_client.do_request(
            "POST",
            f"/v1/chat/{function_path}",
            json={**serialized_data, **kwargs},
        )
        if response.status_code == 429:
            raise RateLimitError("Rate limit error: please retry in a few seconds")
        if response.status_code != 200:
            raise APIError(
                f"Failed to run function {function_path} with status {response.status_code}"
            )
        return FunctionResponse(**response.json())

    async def _chat_stream(
        self, function_path, data: ChatPayload, **kwargs
    ) -> Generator[StreamingChunk, None, None]:
        gen = self.http_client.stream(
            "POST",
            f"/v1/chat/{function_path}",
            json={**data.model_dump(), **kwargs},
            params={"stream": "True"},
        )
        async for item in gen:
            yield StreamingChunk(**item)



from typing import List, Optional

from opperai._http_clients import _http_client
from opperai.types.exceptions import APIError
from opperai.types.indexes import (
    DocumentIn,
    DocumentOut,
    Filter,
    IndexOut,
    IndexRetrieveResponse,
)


class Indexes:
    def __init__(self, http_client: _http_client):
        self.http_client = http_client

    def create(self, name: str) -> int:
        response = self.http_client.do_request(
            "POST",
            "/v1/indexes",
            json={"name": name},
        )
        if response.status_code != 200:
            raise APIError(f"Failed to create index with status {response.status_code}")
        return response.json()["id"]

    def delete(self, index_id: int):
        response = self.http_client.do_request(
            "DELETE",
            f"/v1/indexes/{index_id}",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to delete index with status {response.status_code}")
        return response.json()

    def list(self):
        response = self.http_client.do_request(
            "GET",
            "/v1/indexes",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to list indexes with status {response.status_code}")
        return [IndexOut.model_validate(item) for item in response.json()]

    def get_by_name(self, name: str) -> IndexOut:
        indexes = self.list()
        for index in indexes:
            if index.name == name:
                return index
        return None

    def upload_file(self, index_id: int, file_path: str, **kwargs):
        # Get upload URL
        upload_url_response = self.http_client.do_request(
            "GET",
            f"/v1/indexes/{index_id}/upload_url",
            params={"filename": file_path.split("/")[-1]},
        )
        if upload_url_response.status_code != 200:
            raise APIError(
                f"Failed to get upload URL with status {upload_url_response.status_code}"
            )
        upload_url_data = upload_url_response.json()

        # Upload file
        with open(file_path, "rb") as file:
            files = {"file": (file_path.split("/")[-1], file)}
            print(upload_url_data)
            upload_response = self.http_client.do_request(
                "POST",
                upload_url_data["url"],
                files=files,
                data=upload_url_data["fields"],
            )
            if upload_response.status_code not in [200, 204]:
                raise APIError(
                    f"Failed to upload file with status {upload_response.status_code}"
                )

        # Register file
        register_file_response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{index_id}/register_file",
            json={"uuid": upload_url_data["uuid"], **kwargs},
        )
        if register_file_response.status_code != 200:
            raise APIError(
                f"Failed to register file with status {register_file_response.status_code}"
            )

        return register_file_response.json()

    def index(self, index_id: int, doc: DocumentIn) -> DocumentOut:
        response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{index_id}/index",
            json=doc.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(f"Failed to add document with status {response.status_code}")
        return DocumentOut.model_validate(response.json())

    def retrieve(
        self, index_id: int, query: str, k: int, filters: Optional[List[Filter]] = None
    ) -> List[IndexRetrieveResponse]:
        response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{index_id}/query",
            json={"q": query, "k": k, "filters": filters or []},
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to retrieve index {index_id} with status {response.status_code}"
            )
        return [IndexRetrieveResponse.model_validate(item) for item in response.json()]



import os
from typing import Dict, List, Optional, Union
from unittest.mock import MagicMock, patch

from opperai import fn
from opperai.utils import (
    convert_function_call_to_json,
)
from pydantic import BaseModel


@patch("opperai._http_clients._http_client.do_request")
def test_decorator(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(status_code=404),
        MagicMock(status_code=200, json=lambda: {"id": 1}),
        MagicMock(status_code=200, json=lambda: {"json_payload": "Hola"}),
        MagicMock(status_code=404),
        MagicMock(status_code=200, json=lambda: {"id": 2}),
        MagicMock(status_code=200, json=lambda: {"json_payload": ["Hola", "Bonjour"]}),
    ]

    @fn()
    def translate(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    @fn()
    def translate_list(text: str, target_languages: List[str]) -> List[str]:
        """Translate text to a list of target languages."""

    assert translate("Hello", "es") == "Hola"
    assert translate_list("Hello", ["es", "fr"]) == ["Hola", "Bonjour"]


@patch("opperai._http_clients._http_client.do_request")
def test_decorator_supply_model(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(status_code=404),
        MagicMock(status_code=200, json=lambda: {"id": 1}),
        MagicMock(status_code=200, json=lambda: {"json_payload": "Hola"}),
        MagicMock(status_code=404),
        MagicMock(status_code=200, json=lambda: {"id": 2}),
        MagicMock(status_code=200, json=lambda: {"json_payload": ["Hola", "Bonjour"]}),
    ]

    @fn(model="gpt-4-0125-preview")
    def translate(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    @fn()
    def translate_list(text: str, target_languages: List[str]) -> List[str]:
        """Translate text to a list of target languages."""

    assert translate("Hello", "es") == "Hola"
    assert translate_list("Hello", ["es", "fr"]) == ["Hola", "Bonjour"]


class NestedModel(BaseModel):
    nested_integer: int
    nested_string: str


class ExampleModel(BaseModel):
    integer_field: int
    float_field: float
    string_field: str
    boolean_field: bool
    optional_field: Optional[int]
    list_field: List[int]
    union_field: Union[int, str]
    dict_field: Dict[str, int]
    nested_model: NestedModel
    list_of_models: List[NestedModel]


def test_convert_func_to_json():
    def something(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    json = convert_function_call_to_json(something, "Hello", "es")
    assert json == {"text": "Hello", "target_language": "es"}

    model = ExampleModel(
        integer_field=1,
        float_field=1.0,
        string_field="string",
        boolean_field=True,
        optional_field=1,
        list_field=[1, 2, 3],
        union_field="string",
        dict_field={"key": 1},
        nested_model=NestedModel(nested_integer=1, nested_string="string"),
        list_of_models=[
            NestedModel(nested_integer=1, nested_string="string"),
            NestedModel(nested_integer=1, nested_string="string"),
        ],
    )

    def advanced(text: str, target_language: str, model: ExampleModel) -> str:
        pass

    json = convert_function_call_to_json(advanced, "Hello", "es", model)
    assert json == {
        "text": "Hello",
        "target_language": "es",
        "model": model.model_dump(),
    }



import pytest
import random
from opperai import AsyncClient
from opperai.types import ChatPayload, FunctionDescription, Message, CacheConfiguration


DEFAULT_NAME = "test_function"
DEFAULT_DESCRIPTION = "Test function"
DEFAULT_INSTRUCTIONS = "Do something"


@pytest.fixture
async def function(
    request: pytest.FixtureRequest,
    aclient: AsyncClient,
):
    name = DEFAULT_NAME
    description = DEFAULT_DESCRIPTION
    instructions = DEFAULT_INSTRUCTIONS

    try:
        name = request.param.get("name", DEFAULT_NAME)
        description = request.param.get("description", DEFAULT_DESCRIPTION)
        instructions = request.param.get("instructions", DEFAULT_INSTRUCTIONS)
    except AttributeError:
        pass

    fdesc = FunctionDescription(
        path=f"test/sdk/{name}",
        description=description,
        instructions=instructions,
    )

    return await aclient.functions.create(fdesc)


@pytest.mark.asyncio(scope="module")
async def test_create_function(function, vcr_cassette):
    fid = await function
    assert fid is not None


@pytest.mark.parametrize("function", [{"name": "test_get_by_id"}], indirect=True)
@pytest.mark.asyncio(scope="module")
async def test_get_by_id(aclient: AsyncClient, function, vcr_cassette):
    fid = await function
    f_by_id = await aclient.functions.get_by_id(fid)
    assert f_by_id.path == "test/sdk/test_get_by_id"


@pytest.mark.parametrize("function", [{"name": "test_get_by_path"}], indirect=True)
@pytest.mark.asyncio(scope="module")
async def test_get_by_path(aclient: AsyncClient, function, vcr_cassette):
    await function
    f_by_path = await aclient.functions.get_by_path("test/sdk/test_get_by_path")
    assert f_by_path.path == "test/sdk/test_get_by_path"


@pytest.mark.parametrize("function", [{"name": "test_get"}], indirect=True)
@pytest.mark.asyncio(scope="module")
async def test_get(aclient: AsyncClient, function, vcr_cassette):
    fid = await function
    f_by_id = await aclient.functions.get(id=fid)
    assert f_by_id.path == "test/sdk/test_get"

    f_by_id = await aclient.functions.get(path="test/sdk/test_get")
    assert f_by_id.path == "test/sdk/test_get"


@pytest.mark.parametrize("function", [{"name": "test_update_function"}], indirect=True)
@pytest.mark.asyncio(scope="module")
async def test_update_function(aclient: AsyncClient, function, vcr_cassette):
    fid = await function
    f = await aclient.functions.get_by_id(fid)
    f.instructions = "Do something else"
    fn_id = await aclient.functions.update(f)
    f1 = await aclient.functions.get_by_id(fn_id)

    assert f1.instructions == "Do something else"


@pytest.mark.parametrize(
    "function", [{"name": "test_delete_function_by_id"}], indirect=True
)
@pytest.mark.asyncio(scope="module")
async def test_delete_function_by_id(aclient: AsyncClient, function, vcr_cassette):
    fid = await function
    f = await aclient.functions.get(id=fid)
    assert f is not None
    await aclient.functions.delete(id=f.id)
    f = await aclient.functions.get(id=f.id)
    assert f is None


@pytest.mark.parametrize(
    "function", [{"name": "test_delete_function_by_path"}], indirect=True
)
@pytest.mark.asyncio(scope="module")
async def test_delete_function_by_path(aclient: AsyncClient, function, vcr_cassette):
    fid = await function
    f = await aclient.functions.get(id=fid)
    assert f is not None
    await aclient.functions.delete(path=f.path)
    f = await aclient.functions.get(path=f.path)
    assert f is None


@pytest.mark.parametrize(
    "function",
    [{"name": "test_async_chat", "instructions": "translate to french"}],
    indirect=True,
)
@pytest.mark.asyncio(scope="module")
async def test_async_chat(function, aclient: AsyncClient, vcr_cassette):
    fid = await function
    f = await aclient.functions.get_by_id(fid)
    resp = await aclient.functions.chat(
        f.path, ChatPayload(messages=[Message(role="user", content="hello")])
    )

    assert resp.message == "Bonjour"


@pytest.mark.parametrize(
    "function",
    [{"name": "test_async_chat_stream", "instructions": "translate to french"}],
    indirect=True,
)
@pytest.mark.asyncio(scope="module")
async def test_async_chat_stream(function, aclient: AsyncClient, vcr_cassette):
    fid = await function
    f = await aclient.functions.get_by_id(fid)
    gen = await aclient.functions.chat(
        f.path,
        ChatPayload(messages=[Message(role="user", content="hello")]),
        stream=True,
    )

    resp = "".join([message.delta async for message in gen if message.delta])

    assert resp == "Bonjour"


@pytest.mark.asyncio(scope="module")
async def test_create_function_with_cache(aclient: AsyncClient, vcr_cassette):
    fdesc = FunctionDescription(
        path="test/sdk/test_create_function_with_cache_async",
        description="Test function",
        instructions="Do something",
        cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
    )

    fid = await aclient.functions.create(fdesc)
    assert fid is not None

    res = await aclient.functions.chat(
        fdesc.path, ChatPayload(messages=[Message(role="user", content="hello")])
    )
    print(res)
    assert not res.cached

    res = await aclient.functions.chat(
        fdesc.path, ChatPayload(messages=[Message(role="user", content="hello")])
    )
    print(res)
    assert res.cached

    await aclient.functions.delete(path=fdesc.path)



from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field


class Span(BaseModel):
    uuid: Optional[UUID] = None
    project: Optional[str] = None
    name: Optional[str] = None
    input: Optional[str] = None
    output: Optional[str] = None
    start_time: Optional[datetime] = None
    parent_uuid: Optional[UUID] = None
    end_time: Optional[datetime] = None
    error: Optional[str] = None
    meta: Optional[dict] = None
    evaluations: Optional[dict] = None
    score: Optional[int] = None


class SpanFeedback(BaseModel):
    dimension: Optional[str] = None
    score: Optional[float] = Field(None, ge=0, le=1)
    comment: Optional[str] = None



import os

from ._http_clients import _async_http_client, _http_client
from .functions._async_functions import AsyncFunctions
from .functions._functions import Functions
from .indexes._async_indexes import AsyncIndexes
from .indexes._indexes import Indexes
from .spans._async_spans import AsyncSpans
from .spans._spans import Spans


DEFAULT_API_URL = "https://api.opper.ai"


class AsyncClient:
    _instance = None

    functions: AsyncFunctions
    indexes: AsyncIndexes
    spans: AsyncSpans

    def __new__(
        cls,
        api_key: str = None,
        api_url: str = None,
        default_model: str = None,
        timeout: int = 60,
    ):
        if api_key is None:
            api_key = os.getenv("OPPER_API_KEY")
            if api_key is None:
                raise Exception(
                    "API key is not provided and OPPER_API_KEY environment variable is not set."
                )
        if api_url is None:
            api_url = os.getenv("OPPER_API_URL", DEFAULT_API_URL)
        if default_model is None:
            default_model = os.getenv("OPPER_DEFAULT_MODEL")
        if cls._instance is None:
            cls._instance = super(AsyncClient, cls).__new__(cls)
            cls._instance.http_client = _async_http_client(api_key, api_url, timeout)
            cls._instance.functions = AsyncFunctions(
                cls._instance.http_client, default_model=default_model
            )
            cls._instance.indexes = AsyncIndexes(cls._instance.http_client)
            cls._instance.spans = AsyncSpans(cls._instance.http_client)
            cls._instance.api_key = api_key
            cls._instance.api_url = api_url
        return cls._instance


class Client:
    _instance = None

    functions: Functions
    indexes: Indexes
    spans: Spans

    def __new__(
        cls,
        api_key: str = None,
        api_url: str = None,
        default_model: str = None,
        timeout: int = 60,
    ):
        if api_key is None:
            api_key = os.getenv("OPPER_API_KEY")
            if api_key is None:
                raise Exception(
                    "API key is not provided and OPPER_API_KEY environment variable is not set."
                )
        if api_url is None:
            api_url = os.getenv("OPPER_API_URL", DEFAULT_API_URL)
        if default_model is None:
            default_model = os.getenv("OPPER_DEFAULT_MODEL")
        if cls._instance is None:
            instance = super(Client, cls).__new__(cls)
            instance.http_client = _http_client(api_key, api_url, timeout)
            instance.functions = Functions(
                instance.http_client, default_model=default_model
            )
            instance.indexes = Indexes(instance.http_client)
            instance.spans = Spans(instance.http_client)
            instance.api_key = api_key
            instance.api_url = api_url
            cls._instance = instance
        return cls._instance



import json
from uuid import UUID

from opperai._http_clients import _async_http_client
from opperai.types.spans import Span, SpanFeedback
from opperai.types.exceptions import APIError
from opperai.utils import DateTimeEncoder


class AsyncSpans:
    def __init__(self, http_client: _async_http_client):
        self.http_client = http_client

    async def create(self, span: Span, **kwargs) -> str:
        span_data = span.model_dump(exclude_none=True)
        json_payload = json.dumps(span_data, cls=DateTimeEncoder)
        response = await self.http_client.do_request(
            "POST",
            "/v1/spans",
            data=json_payload,
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to create span {span.name} with status {response.status_code}"
            )

        return response.json()["uuid"]

    async def update(self, span_uuid: UUID, **kwargs) -> str:
        span = Span(uuid=span_uuid, **kwargs)
        json_payload = json.dumps(
            span.model_dump(exclude_none=True), cls=DateTimeEncoder
        )
        response = await self.http_client.do_request(
            "PUT",
            f"/v1/spans/{span.uuid}",
            data=json_payload,
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to update span `{span.name}` with status {response.status_code}"
            )

        return response.json()["uuid"]

    async def delete(self, span_uuid: UUID) -> bool:
        response = await self.http_client.do_request(
            "DELETE",
            f"/v1/spans/{span_uuid}",
        )
        if response.status_code != 204:
            raise APIError(
                f"Failed to update span `{span_uuid}` with status {response.status_code}"
            )

        return True

    async def save_example(self, uuid: str, **kwargs) -> str:
        response = await self.http_client.do_request(
            "POST",
            f"/v1/spans/{uuid}/save_examples",
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to save examples for span {uuid} with status {response.status_code}"
            )

        return response.json()["uuid"]

    async def save_feedback(self, uuid: str, feedback: SpanFeedback, **kwargs) -> str:
        response = await self.http_client.do_request(
            "POST",
            f"/v1/spans/{uuid}/feedbacks",
            json=feedback.model_dump(exclude_unset=True),
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to add feedback for span {uuid} with status {response.status_code}"
            )

        return response.json()
